#ifndef FrameCPP_VERSION_8_FrVect_HH
#define FrameCPP_VERSION_8_FrVect_HH

#include <memory>
#include <vector>
#include <string>

#include "general/types.hh"
#include "framecpp/FrameCPP.hh"

#include "framecpp/Common/Compression.hh"
#include "framecpp/Common/FrVect.hh"

#include "framecpp/Version7/FrVect.hh"

#include "framecpp/Version8/FrameSpec.hh"
#include "framecpp/Version8/Dimension.hh"
#include "framecpp/Version8/GPSTime.hh"
#include "framecpp/Version8/STRING.hh"

namespace FrameCPP
{
  namespace Version_8
  {
    using Previous::FrVectDataTypes;

    class FrVect
      : public ObjectWithChecksum< Common::FrameSpec::Object >,
        public Common::FrVect,
        public FrVectDataTypes
    {
    public:
      // using FrameCPP::Compression;

      enum compression_scheme_type {
	//---------------------------------------------------------
	//---------------------------------------------------------

#if WORDS_BIGENDIAN
#define NATIVE_DECL(x) x = BIGENDIAN_##x
#else
#define NATIVE_DECL(x) x = LITTLEENDIAN_##x
#endif
#define CST_DECL(x,y) BIGENDIAN_##x = y, LITTLEENDIAN_##x = y + 256, NATIVE_DECL(x)

	CST_DECL(RAW,0),
	CST_DECL(GZIP,1),
        CST_DECL(DIFF_GZIP,3),
	CST_DECL(ZERO_SUPPRESS_WORD_2,5),
	CST_DECL(ZERO_SUPPRESS_WORD_4,8),
	CST_DECL(ZERO_SUPPRESS_WORD_8,10),

#undef CST_DECL
#undef NATIVE_DECL

	//---------------------------------------------------------
	// Declaration of meta modes
	//---------------------------------------------------------
	ZERO_SUPPRESS_OTHERWISE_GZIP = 6,
	BEST_COMPRESSION = 9,

	//---------------------------------------------------------
	// Backwards compatability
	//---------------------------------------------------------
	ZERO_SUPPRESS_SHORT = ZERO_SUPPRESS_WORD_2,
	ZERO_SUPPRESS_INT_FLOAT = ZERO_SUPPRESS_WORD_4,
#undef GENERIC
	NONE = RAW
      };

      enum data_types_type {
	FR_VECT_C = 0,
	FR_VECT_2S = 1,
	FR_VECT_8R = 2,
	FR_VECT_4R = 3,
	FR_VECT_4S = 4,
	FR_VECT_8S = 5,
	FR_VECT_8C = 6,
	FR_VECT_16C = 7,
	FR_VECT_STRING = 8,
	FR_VECT_2U = 9,
	FR_VECT_4U = 10,
	FR_VECT_8U = 11,
	FR_VECT_1U = 12
      };


      typedef INT_2U compress_type;
      typedef INT_2U type_type;
      typedef INT_4U nData_type;
      typedef INT_4U nBytes_type;
      typedef INT_4U nDim_type;

      static const int DEFAULT_GZIP_LEVEL;

      static std::auto_ptr< Object >
      Promote( INT_2U Source,
	       Object* Obj, Common::IStream* Stream );

      static const char* const StructName( );

      static const Common::Description* StructDescription( );

      //: Default constructor
      FrVect( );

      //-----------------------------------------------------------------
      //
      //: Copy Constructor.  
      //
      //!param: const Vect& vect - Copied vector.
      //
      //!exc: std::bad_alloc - Memory allocation failed.
      //
      FrVect( const FrVect& vect );


      //-----------------------------------------------------------------
      //
      //: Constructor.
      //
      //!param: const std::string& name - The name of the data.
      //!param: INT_2U type - The data type.
      //!param: INT_4U nDim - The number of dimensions.
      //!param: const Dimension* dims - A pointer to 'nDim' Dimension objects
      //+       containing information about the dimensions for this data.
      //!param: const byte_order_type byte_order - Byte order of the data. Default
      //+       is BIGENDIAN.   
      //!param: const void* data - A pointer to the data.  Default: 0
      //!param: const std::string& unitY - Units for the data.  Default: ""
      //
      //!exc: std::bad_alloc - Memory allocation failed.
      //    
      FrVect( const std::string& name, INT_2U type, INT_4U nDim,
	      const Dimension* dims,
	      const byte_order_type byte_order = BYTE_ORDER_HOST,
	      const void* data = 0, 
	      const std::string& unitY = "" );

      //-----------------------------------------------------------------
      //
      //: Constructor.
      //
      //!param: const std::string& name - The name of the data.
      //!param: INT_2U type - The data type.
      //!param: INT_4U nDim - The number of dimensions.
      //!param: const Dimension* dims - A pointer to 'nDim' Dimension objects
      //+     containing information about the dimensions for this data.
      //!param: const byte_order_type byte_order - Byte order of the data.
      //+	Default is BIGENDIAN.      
      //!param: const void* data - A pointer to the data.  Default: 0
      //!param: const std::string& unitY - Units for the data.  Default: ""
      //!param: bool allocate - Whether to allocate data. Default: true.
      //!param: bool owns - Whether to destruct data in the destructor.
      //+	Default: true.
      //
      //!exc: std::bad_alloc - Memory allocation failed.
      //
      FrVect( const std::string& name, INT_2U type, INT_4U nDim,
	      const Dimension* dims,
	      const byte_order_type byte_order = BYTE_ORDER_HOST,
	      void* data = 0,
	      const std::string& unitY = "", 
	      bool allocate = true, bool owns = true );

      FrVect( const std::string& name,
	      compress_type Compress,
	      type_type type,
	      nDim_type nDim,
	      const Dimension* dims,
	      nData_type NData,
	      nBytes_type NBytes,
	      void* data,
	      const std::string& unitY = "", 
	      bool allocate = true,
	      bool owns = true );

      //!exc: std::bad_alloc - Memory allocation failed.   
      template< class T >
      FrVect( const std::string& name, INT_4U nDim = 1,
	      const Dimension* dims = 0,
	      const T* data = 0, const std::string& unitY = "" );

      //!exc: std::bad_alloc - Memory allocation failed.   
      template< class T >
      FrVect( const std::string& name, INT_4U nDim = 1,
	      const Dimension* dims = 0,
	      T* data = 0, const std::string& unitY = "",
	      bool allocate = true, bool owns = true );

      FrVect( const Previous::FrVect& Source,
	      Common::IStream* Stream );

      virtual ~FrVect( );

      virtual const char* const ObjectStructName( ) const;

      //: Comparison operator
      bool operator==( const FrVect& RHS ) const;
      bool operator!=( const FrVect& RHS ) const;
      //: Concatination operator
      FrVect& operator+=( const FrVect& RHS );

      //-----------------------------------------------------------------
      /// \brief comparison operator
      //-----------------------------------------------------------------
      virtual bool operator==( const Common::FrameSpec::Object& Obj ) const;

      // Accessors
      const std::string& GetName( ) const;
      INT_2U GetCompress( ) const;
      INT_2U GetType( ) const;
      INT_8U GetNData( ) const;
      INT_8U GetNBytes( ) const;
      const CHAR_U* GetData( ) const;
      const CHAR_U* GetDataRaw( ) const;
      CHAR_U* GetDataRaw( );
      CHAR_U* GetDataUncompressed( );
      const CHAR_U* GetDataUncompressed( General::AutoArray< CHAR_U >&
					 Expanded ) const;
      template< class T > static INT_2U GetDataType();
      INT_4U GetNDim( ) const;
      const Dimension& GetDim( INT_4U Offset ) const;
      Dimension& GetDim( INT_4U Offset );
      const std::string GetUnitY( ) const;
      CHAR_U* SteelData( bool& Owns ) const;

      // Mutators
      void Compress( compression_scheme_type Scheme, int GZipLevel );
      void Uncompress( );

      virtual void CompressData( INT_4U Scheme, INT_2U GZipLevel );

      virtual Compression::compression_base_type
      Compression( ) const;

      virtual Common::FrameSpec::Object*
      CloneCompressed( cmn_compression_scheme_type Scheme,
		       cmn_compression_level_type Level ) const;

      //-----------------------------------------------------------------
      /// \brief Establish the channel name.
      ///
      /// \param[in] Name
      ///     The channel name
      //-----------------------------------------------------------------
      void SetName( const std::string& Name );

      void SetNData( INT_4U NData );

      size_t GetTypeSize( ) const;
      static size_t GetTypeSize( INT_2U type );

      //-----------------------------------------------------------------
      /// \brief Merge with another FrAdcData
      //-----------------------------------------------------------------
      FrVect& Merge( const FrVect& RHS );

      std::auto_ptr< FrVect > SubFrVect( INT_4U Start, INT_4U Stop ) const;

    protected:
      //-----------------------------------------------------------------
      /// \brief Down grade an object
      //-----------------------------------------------------------------
      virtual std::auto_ptr< Object >
      demote( INT_2U Target, Common::IStream* Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Upgrade an object
      //-----------------------------------------------------------------
      virtual std::auto_ptr< Object >
      promote( INT_2U Target, Object* Obj, Common::IStream* Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Number of bytes needed to write this structure
      //-----------------------------------------------------------------
      virtual FrameCPP::cmn_streamsize_type
      pBytes( const Common::StreamBase& Stream ) const;

      virtual FrVect* pCreate( Common::IStream& Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Verify the integrity of the object
      ///
      /// \param Verifier
      ///     The handle to the Verify object used to store data that
      ///     needs to be accumulated before being processed.
      /// \param Stream
      ///     The input stream from which the object was created.
      ///
      /// \remarks
      ///     Upon failure, this method thows an exception.
      //-----------------------------------------------------------------
      virtual void VerifyObject( Common::Verify& Verifier,
				 Common::IFrameStream& Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Write the structure to the stream
      //-----------------------------------------------------------------
      virtual void pWrite( Common::OStream& Stream ) const;

      void
      expandToBuffer( General::AutoArray< CHAR_U >& Dest,
		      INT_8U& DestNBytes ) const;

    private:
      //-----------------------------------------------------------------
      //-----------------------------------------------------------------
      typedef ObjectWithChecksum< Common::FrameSpec::Object > object_type;

      //-----------------------------------------------------------------
      /// \brief Dictionary class id for this structure.
      //-----------------------------------------------------------------
      static const Common::FrameSpec::Info::frame_object_types s_object_id
      = Common::FrameSpec::Info::FSI_FR_VECT;
      
      using Common::FrameSpec::Object::Create;

      struct data_definition_type
      {
      public:

	STRING				name;
	INT_2U				compress;
	INT_2U				type;
	INT_8U				nData;
	INT_8U				nBytes;
	CHAR_U*				data;
	std::vector< Dimension >	dimension;
	STRING				unitY;

        data_definition_type();
        data_definition_type( const std::string& n, INT_2U t, 
			      const byte_order_type byte_order, INT_4U ndim,
			      const Dimension* dims, const std::string& unit );

	void copy_core( const data_definition_type& Source );

	bool operator==( const data_definition_type& RHS ) const;

      };
 
      data_definition_type	m_data;
      bool			m_owns;

      FrVect( Common::IStream& Stream );

      void copy_data( CHAR_U* Data, INT_8U size,
		      bool allocate, bool own );

    };


    inline Compression::compression_base_type FrVect::
    Compression( ) const
    {
      return ( 0x0FF & m_data.compress );
    }

    inline std::auto_ptr< Common::FrameSpec::Object > FrVect::
    Promote( INT_2U Source,
	     Common::FrameSpec::Object* Obj, Common::IStream* Stream )
    {
      return Object::PromoteObject< Previous::FrVect, FrVect >
	( DATA_FORMAT_VERSION, Source, Obj, Stream );
    }

    inline const char* const FrVect::
    StructName( )
    {
      static const CHAR* class_name( "FrVect" );
      return class_name;
    }

    //-------------------------------------------------------------------
    /// Return a pointer to the data being managed by the FrVect
    /// structure.
    ///
    /// \see GetDataRaw, GetDataUncompressed
    //-------------------------------------------------------------------
    inline const CHAR_U* FrVect::
    GetData( ) const
    {
      return GetDataRaw( );
    }

    inline INT_2U FrVect::
    GetCompress( ) const
    {
      return m_data.compress;
    }

    inline INT_2U FrVect::
    GetType( ) const
    {
      return m_data.type;
    }

    inline size_t FrVect::
    GetTypeSize( ) const
    {
      return GetTypeSize( m_data.type );
    }

    inline INT_8U FrVect::
    GetNBytes( ) const
    {
      return m_data.nBytes;
    }

    inline INT_4U FrVect::GetNDim( ) const
    {
      return m_data.dimension.size( );
    }

    inline const Dimension& FrVect::
    GetDim( INT_4U Offset ) const
    {
      //:TODO: Throw exception if value out of range
      return m_data.dimension[ Offset ];
    }

    inline Dimension& FrVect::
    GetDim( INT_4U Offset )
    {
      //:TODO: Throw exception if value out of range
      return m_data.dimension[ Offset ];
    }

    inline const std::string FrVect::
    GetUnitY( ) const
    {
      return m_data.unitY;
    }

    //-------------------------------------------------------------------
    /// Assign a new value for the channel name.
    //-------------------------------------------------------------------
    inline void FrVect::
    SetName( const std::string& Name )
    {
      m_data.name = Name;
    }

    inline void FrVect::
    SetNData( INT_4U NData )
    {
      m_data.nData = NData;
    }
   
    inline CHAR_U* FrVect::
    SteelData( bool& Owns ) const
    {
      Owns = m_owns;
      const_cast< FrVect* >( this )->Uncompress( );
      const_cast< FrVect* >( this )->m_owns = false;
      return m_data.data;
    }
   
  } // namespace - Version_8
} // namespace FrameCPP

#if defined( __SUNPRO_CC ) && ( __SUNPRO_CC <= 0x550 )
#include "framecpp/Version8/FrVect.icc"
#endif /* */

#endif /* FrameCPP_VERSION_8_FrVect_HH */
